import asyncio
import logging
from typing import Optional

import pendulum
import schedule
from komoutils.core import KomoBase, safe_ensure_future

from aporacle.intervaled_data_symbol.client.cursor import Cursor
from aporacle.intervaled_data_symbol.client.intervaled_symbol import IntervaledSymbol
from aporacle.intervaled_data_symbol.client.shared_data import SharedData


class Finalizer(KomoBase):

    def __init__(self, cursor: Cursor):
        self.cursor: Cursor = cursor
        schedule.every(1).seconds.do(self.run)

    async def start(self):
        self.log_with_clock(log_level=logging.INFO, msg=f"Initializing orchestrator summarizer. ")
        while True:
            schedule.run_pending()
            await asyncio.sleep(1)
            continue

    def run(self):
        self.cursor.finalizer(stamp=int(pendulum.now().subtract(seconds=2).timestamp()))


class Orchestrator(KomoBase):

    def __init__(self,
                 symbol: str,
                 input_queue: asyncio.Queue,
                 output_queue: asyncio.Queue,
                 output_record_evaluation_ticks: int = 18
                 ):
        self.symbol: str = symbol
        self.input_queue: asyncio.Queue = input_queue
        self.output_queue: asyncio.Queue = output_queue
        self.output_record_evaluation_ticks = output_record_evaluation_ticks

        self.intervaled: Optional[IntervaledSymbol] = None
        self.finalizer: Optional[Finalizer] = None

        self.initialize()

    def initialize(self):
        self.intervaled = IntervaledSymbol(symbol=self.symbol,
                                           input_trading_data_queue=self.input_queue,
                                           output_record_evaluation_ticks=self.output_record_evaluation_ticks)
        self.finalizer = Finalizer(cursor=self.intervaled.cursor)

    def start(self):
        self.intervaled.start()
        safe_ensure_future(self.finalizer.start())

    def tso_price_set(self, chain: str, tso: str, price: float):
        if self.symbol.split("_")[-2] not in [tso]:
            self.log_with_clock(log_level=logging.WARNING, msg=f"Incorrect TSO for price set. Tried to set {tso} but "
                                                               f"{self.symbol.split('_')[-2]} is allowable. ")
            return

        SharedData.get_instance().current_tso_prices[chain] = {tso: price}

    def generate_chain_epoch_record(self, chain: str, epoch: int):
        prices, returns = self.intervaled.run(chain=chain, epoch=epoch)
        safe_ensure_future(self.output_queue.put({"asset": self.symbol,
                                                  "chain": chain,
                                                  "epoch": epoch,
                                                  "prices": prices,
                                                  "returns": returns
                                                  }))



import asyncio
import time
from collections import OrderedDict
from typing import Optional

from aporacle.intervaled_data_symbol.client.symbol import Symbol
from aporacle.intervaled_data_symbol.client.tick import Tick


class IntervaledSymbol(Symbol):
    def __init__(self,
                 symbol: str,
                 input_trading_data_queue: asyncio.Queue,
                 output_record_evaluation_ticks: int = 15):
        super().__init__(symbol, input_trading_data_queue)
        self.output_record_evaluation_ticks = output_record_evaluation_ticks
        self.evaluation_ticks: Optional[OrderedDict[str, Tick]] = None

    def create_prices_output_record(self, chain: str, epoch: int):
        prices = {"chain": chain, "epoch": epoch}
        for index, tick in enumerate(reversed(self.evaluation_ticks.values())):
            prices.update(tick.get_prices(index=index))

        sma_ticks = list(reversed(self.evaluation_ticks.values()))[0:5]
        for index, tick in enumerate(sma_ticks):
            prices.update(tick.get_prices_smas(index=index))

        return prices

    def create_returns_output_record(self, chain: str, epoch: int):
        returns = {"chain": chain, "epoch": epoch}

        for index, tick in enumerate(reversed(self.evaluation_ticks.values())):
            returns.update(tick.get_returns(chain=chain, index=index))

        sma_ticks = list(reversed(self.evaluation_ticks.values()))[0:5]
        for index, tick in enumerate(sma_ticks):
            returns.update(tick.get_returns_smas(chain=chain, index=index))

        return returns

    def run(self, chain: str, epoch: int):
        try:
            t1 = time.perf_counter()
            self.evaluation_ticks = self.cursor.evaluation_ticks(age=self.output_record_evaluation_ticks)
            prices = self.create_prices_output_record(chain=chain, epoch=epoch)
            returns = self.create_returns_output_record(chain=chain, epoch=epoch)
            t2 = time.perf_counter()
            self.log_with_clock(log_level=20, msg=f"Chain {chain} epoch {epoch} took {t2 - t1} seconds. ")
            return prices, returns
        except Exception as e:
            return 'failure'



import logging
import os

import requests
from komoutils.core.time import the_time_in_iso_now_is

from aporacle.intervaled_data_symbol.execution.executor import IntervaledTradingSymbolExecutor

if "TSO_SETUP_URL" not in os.environ:
    raise Exception("TSO_SETUP_URL value MUST be set. ")

logging.basicConfig(level=logging.INFO)

app_logger = logging.getLogger(__name__)
app_logger.setLevel(logging.INFO)


def log_with_clock(log_level: int, msg: str, **kwargs):
    app_logger.log(log_level, f"{msg} [clock={str(the_time_in_iso_now_is())}]", **kwargs)


def get_api_setup_data():
    url = f"http://{os.getenv('TSO_SETUP_URL')}/setup"

    try:
        with requests.Session() as session:
            response = session.get(url=url)
            response.raise_for_status()
            setup = response.json()
            return setup

    except requests.exceptions.HTTPError as e:
        app_logger.error(f"Please ensure that {url} is valid")
    except requests.exceptions.ConnectionError as e:
        raise SetupResourceNotAvailableException()
    except Exception as e:
        raise SetupResourceNotAvailableException(f"Undocumented asset setup acquisition error - {e}")


class SetupResourceNotAvailableException(Exception):
    """Exception raised when a connection error happens while accessing TSO_SETUP.

        Attributes:
            message -- explanation of the error
        """

    def __init__(self, message: str = "TSO Setup not available. "):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message}'


class ConversionRateResourceNotAvailableException(Exception):
    """Exception raised when a connection error happens while accessing TSO_SETUP.

        Attributes:
            message -- explanation of the error
        """

    def __init__(self, message: str = "Conversion rate resource error. "):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message}'


class SymbolNotFoundException(Exception):
    def __init__(self, symbol: str):
        self.symbol = symbol


class SymbolHasNoRateException(Exception):
    def __init__(self, symbol: str):
        self.symbol = symbol


__all__ = [
    "IntervaledTradingSymbolExecutor"
]



from typing import Optional


class SharedData:
    _shared_instance: "SharedData" = None

    @classmethod
    def get_instance(cls) -> "SharedData":
        if cls._shared_instance is None:
            cls._shared_instance = SharedData()
        return cls._shared_instance

    def __init__(self):
        self.current_chain_epoch: dict = {}
        self.current_tso_prices: dict = {"songbird": float, "flare": float}

        self.last_close: Optional[float] = None
        self.last_returns: dict = {"flare": 0.0, "songbird": 0.0}

    # def set(self, epoch_data: dict):
    #     end_time = epoch_data["end_time"] - SharedSetup.get_instance().end_time_execution_buffer - 1
    #     start_time = epoch_data["start_time"] - SharedSetup.get_instance().end_time_execution_buffer
    #
    #     self.current_chain_epoch[epoch_data["chain"]] = {
    #         "epoch": epoch_data["epoch_id"],
    #         "start": start_time,
    #         "end": end_time,
    #         "st_str": epoch_data["st_str"],
    #         "et_str": epoch_data["et_str"],
    #     }



import asyncio
import enum
import logging
from collections import OrderedDict

import pendulum
import schedule
from komoutils.core import KomoBase
from komoutils.core.time import the_time_in_iso_now_is

from aporacle.intervaled_data_symbol.client.shared_data import SharedData
from aporacle.intervaled_data_symbol.client.tick import Tick


class CursorStatus(enum.Enum):
    STARTED = "STARTED"
    STOPPED = "STOPPED"
    PENDING = "PENDING"


class Cursor(KomoBase):

    def __init__(self, symbol: str):
        self.symbol: str = symbol
        self.tso: str = str(symbol).split("_")[2]
        self.quote: str = str(symbol).split("_")[3]
        self.ticks: OrderedDict = OrderedDict()

        self.status: CursorStatus = CursorStatus.PENDING

    def evaluation_ticks(self, age: int = 15) -> OrderedDict[str, Tick]:
        new_ticks = OrderedDict()
        for tick in self.ticks.values():
            if tick.age < age and tick.finalized is True:
                new_ticks[tick.stamp] = tick

        return new_ticks

    async def start(self):
        self.log_with_clock(log_level=logging.INFO, msg=f"Starting cursor for {self.symbol}. ")
        self.generator()
        # self.status = CursorStatus.STARTED
        schedule.every(5).seconds.do(self.generator)
        schedule.every(60).seconds.do(self.destroyer)
        while True:
            schedule.run_pending()
            await asyncio.sleep(1)
            if self.status == CursorStatus.PENDING:
                self.log_with_clock(log_level=logging.INFO,
                                    msg=f"Cursor for {self.symbol} has a status of {self.status.value}. ")
            continue

    def generator(self):
        stamp = int(pendulum.now().timestamp())
        for s in range(stamp, stamp + 15, 1):
            if s in self.ticks:
                continue

            self.ticks[s] = Tick(symbol=self.symbol, stamp=s)

        finalized = [t for t in self.ticks.values() if t.finalized]
        self.log_with_clock(log_level=logging.INFO, msg=f"{len(finalized)} of {len(self.ticks)} of ticks for "
                                                        f"{self.symbol} have been finalized. ")

    def destroyer(self):
        candidates = [stamp for stamp, tick in self.ticks.items() if tick.expires_at < pendulum.now().timestamp()]
        for stamp in candidates:
            self.ticks.pop(stamp)

        self.log_with_clock(log_level=logging.INFO, msg=f"Destroyed {len(candidates)} ticks for {self.symbol} at "
                                                        f"{the_time_in_iso_now_is()}. ")

    def finalizer(self, stamp: int):
        try:
            # print(self.status)
            if self.status != CursorStatus.STARTED:
                return

            if len(self.ticks) == 0:
                self.log_with_clock(log_level=logging.WARNING, msg=f"No ticks available for {self.symbol}. ")

            assert stamp in self.ticks, f"Stamp error on {self.symbol}. "

            # print(f"Finalizing {stamp} at {the_time_in_iso_now_is()}.")

            sorted_dict = dict(sorted(self.ticks.items())).copy()
            means = [tick.mean for tick in sorted_dict.values() if tick.age < 120 and tick.finalized is True]
            returns: dict = {"flare": [], "songbird": []}
            try:
                for chain in ['flare', 'songbird']:
                    returns[chain] = [tick.returns[chain]
                                      for tick in sorted_dict.values()
                                      if tick.age < 120 and tick.finalized is True and tick.returns[chain] is not None]
            except Exception as e:
                pass

            # for tick in sorted_dict.values():
            #     print(f"{tick.stamp} {tick.returns}")


            # print(means)
            self.ticks[stamp].finalize_prices(previous_price_means=means)
            self.ticks[stamp].finalize_returns(previous_returns_means=returns)

            SharedData.get_instance().last_close = self.ticks[stamp].close
            SharedData.get_instance().last_returns = self.ticks[stamp].returns
            # self.ticks[stamp + 1].initialize(previous_close=self.ticks[stamp].close,
            #                                  previous_returns=self.ticks[stamp].returns)

        except AssertionError as ae:
            self.log_with_clock(log_level=logging.ERROR, msg=f"{ae}")
        except Exception as e:
            raise



import logging

import numpy as np
import pendulum
from komoutils.core import KomoBase

from aporacle.intervaled_data_symbol.client.shared_data import SharedData
from aporacle.intervaled_data_symbol.client.trades import Trades


class Tick(KomoBase):

    def __init__(self, symbol: str, stamp: int, ttl: int = 180):
        self.symbol: str = symbol
        self.tso: str = str(symbol).split("_")[2]
        self.quote: str = str(symbol).split("_")[3]

        self.trades: list = []
        self.stamp: int = stamp
        self.ttl: int = ttl
        self.expires_at: int = self.stamp + self.ttl
        self.time: str = pendulum.from_timestamp(self.stamp).to_datetime_string()
        self.open: float = 0
        self.low: float = 0
        self.high: float = 0
        self.close: float = 0
        self.mean: float = 0
        self.volume: float = 0
        self.count: float = 0
        self.elapsed: float = 0

        self.previous_price_means: list = []
        self.previous_returns_means: dict = {"flare": [], "songbird": []}
        self.prices_smas: dict = {}
        self.returns_smas: dict = {}
        self.emas: dict = {}
        # Define window size for the moving average
        self.window_sizes = [5, 10, 15, 20, 30][0:5]

        # self.returns: float = 0
        self.returns: dict = {"flare": 0.0, "songbird": 0.0}  # {chain: return}
        self.finalized: bool = False

    @property
    def age(self):
        return pendulum.from_timestamp(self.stamp).diff(pendulum.now()).in_seconds()

    @property
    def summary(self) -> dict:
        base = {
            'base': self.tso,
            'quote': self.quote,
            'stamp': self.stamp,
            'time': self.time,
            'elapsed': self.elapsed,
            'mean_feature': round(self.mean, 8),
            'return_flare_feature': round(self.returns['flare'], 8),
            'return_songbird_feature': round(self.returns['songbird'], 8),
            'volume_feature': self.volume,
            'count': self.count,
        }

        base.update(self.prices_smas)
        base.update(self.emas)

        return base

    def initialize(self):
        # assert previous_close > 0, f"Previous close of ZERO or less encountered. {self.executor} {self.stamp}"
        previous_close = SharedData.get_instance().last_close
        previous_returns = SharedData.get_instance().last_returns
        self.mean = previous_close
        self.open = previous_close
        self.low = previous_close
        self.high = previous_close
        self.close = previous_close
        # self.returns = previous_returns
        self.volume = 1e-25
        self.count = 0
        self.elapsed = 0

    def summarize(self):
        p = [SharedData.get_instance().last_close] + [item['price'] for item in self.trades]
        prices = np.array(p)
        self.open = prices.flat[0]
        self.low = prices.min()
        self.high = prices.max()
        self.close = prices.flat[-1]
        self.mean = prices.mean()
        self.volume = np.array([item['amount'] for item in self.trades]).sum()
        self.count = len(prices)
        print(f"Count: {self.count} Volume: {self.volume}")

    def get_prices_smas(self, index: int):
        return {f"{key}_{index}": value for key, value in self.prices_smas.items()}

    def get_returns_smas(self, chain: str, index: int):
        chain_returns: dict = self.returns_smas[chain]
        return {f"{key}_{index}": value for key, value in chain_returns.items()}

    def get_prices(self, index: int):
        return {f"{key}_{index}": value for key, value in self.summary.items() if 'feature' in key and 'mean' in key}

    def get_returns(self, chain: str, index: int):
        return {f"{key}_{index}": value for key, value in self.summary.items() if
                'feature' in key and 'return' in key and chain in key}

    def simple_moving_average(self, metric: str):
        # start_time = time.perf_counter()

        if metric == 'prices':
            # Create a numpy array from the values
            means = np.array([previous_mean for previous_mean in self.previous_price_means] + [self.mean])
            # print(f"Tick means: {means}")
            # Calculate the simple moving average
            # We use np.convolve for moving average calculation
            # 'valid' ensures that only fully overlapping sections are used
            # dividing by window_size gives the average
            for window_size in self.window_sizes:
                self.prices_smas[f"sma_{window_size}secs_{metric}_feature"] \
                    = np.convolve(means, np.ones(window_size) / window_size, mode='valid')[-1]

        elif metric == 'returns':
            for chain in ['flare', 'songbird']:
                # Create a numpy array from the values
                # print(f"Previous returns: {self.previous_returns_means}")
                # print(f"Current returns: {self.returns}")
                returns \
                    = np.array([previous_returns
                                for previous_returns in self.previous_returns_means[chain]] + [self.returns[chain]])
                # print(returns)
                _smas = {}
                for window_size in self.window_sizes:
                    _smas[f"sma_{window_size}secs_{metric}_{chain}_feature"] \
                        = np.convolve(returns, np.ones(window_size) / window_size, mode='valid')[-1]

                self.returns_smas[chain] = _smas

    def exponential_moving_average(self, weighting_factor=0.2, metric: str = 'returns'):
        # start_time = time.perf_counter()

        means = np.array([previous_mean for previous_mean in self.previous_price_means] + [self.mean])

        for period in self.window_sizes:
            ema = np.zeros(len(means))
            sma = np.mean(means[:period])
            ema[period - 1] = sma
            for i in range(period, len(means)):
                ema[i] = (means[i] * weighting_factor) + (ema[i - 1] * (1 - weighting_factor))

            self.emas[f"ema_{period}secs_{metric}_feature"] = ema.tolist()[-1]

        # end_time = time.perf_counter()

        # print(f"Time elapsed {end_time - start_time}")

    def finalize_prices(self, previous_price_means: list):
        try:
            # If there are no trades, no summaries must happen. The initialized figures stand.
            self.trades.clear()
            if self.stamp in Trades.get_instance().trades:
                self.trades = Trades.get_instance().trades[self.stamp]

            if len(self.trades) == 0:
                self.initialize()

            print(f"Finalizing found {len(self.trades)} trades for {self.symbol} at {self.time}. ")

            # Collect previous tick means
            self.previous_price_means = previous_price_means
            # Summarize the tick trades
            self.summarize()
            self.simple_moving_average(metric='prices')

        except AssertionError as ae:
            if self.close == 0:
                self.log_with_clock(log_level=logging.WARNING,
                                    msg=f"{ae}. {self.close}. Will update with aggregated quote price when available. ")
            else:
                print(f"Error {ae}")

        except Exception as e:
            print(f"Error in finalizing {self.symbol} at {self.time}. {e}")
            raise

        self.finalized = True

    def finalize_returns(self, previous_returns_means=None):
        if previous_returns_means is None:
            previous_returns_means = {"flare": [], "songbird": []}

        try:
            # If there are no trades, no summaries must happen. The initialized figures stand.
            self.trades.clear()
            if self.stamp in Trades.get_instance().trades:
                self.trades = Trades.get_instance().trades[self.stamp]

            if len(self.trades) == 0:
                self.initialize()

            # Collect previous tick means
            self.previous_returns_means = previous_returns_means
            # calculate returns
            current_price = {}
            for chain in ['flare', 'songbird']:
                current_price[chain] = SharedData.get_instance().current_tso_prices[chain][self.tso]
                amt = (self.mean - current_price[chain]) / abs(current_price[chain])
                self.returns[chain] = amt

            self.simple_moving_average(metric='returns')

        except AssertionError as ae:
            if self.close == 0:
                self.log_with_clock(log_level=logging.WARNING,
                                    msg=f"{ae}. {self.close}. Will update with aggregated quote price when available. ")
            else:
                print(f"Error {ae}")

        except Exception as e:
            print(f"Error in finalizing {self.symbol} at {self.time}. {e}")
            raise

        self.finalized = True



from typing import Dict


class Trades:
    _shared_instance: "Trades" = None

    @classmethod
    def get_instance(cls) -> "Trades":
        if cls._shared_instance is None:
            cls._shared_instance = Trades()
        return cls._shared_instance

    def __init__(self):
        self.trades: Dict[int, list] = {}  # { stamp: list(trades) }



import asyncio
import logging

import pendulum
from komoutils.core import safe_ensure_future, KomoBase

from aporacle.intervaled_data_symbol.client.cursor import Cursor, CursorStatus
from aporacle.intervaled_data_symbol.client.shared_data import SharedData
from aporacle.intervaled_data_symbol.client.tick import Tick
from aporacle.intervaled_data_symbol.client.trades import Trades


class Symbol(KomoBase):

    def __init__(self, symbol: str, input_trading_data_queue: asyncio.Queue):
        self.symbol = symbol
        self.input_trading_data_queue: asyncio.Queue = input_trading_data_queue

        self.tso = str(self.symbol).split("_")[2]
        self.cursor: Cursor = Cursor(symbol=self.symbol)

    def start(self):
        safe_ensure_future(self.cursor.start())
        safe_ensure_future(self.trades_data_loop())

    async def trades_data_loop(self):
        self.log_with_clock(log_level=logging.INFO, msg=f"Started trades data loop for {self.symbol}. ")

        try:
            self.log_with_clock(log_level=logging.INFO, msg=f"Waiting for the first trade for {self.symbol}. ")
            while True:
                trade = await self.input_trading_data_queue.get()
                if 'type' not in trade:
                    continue

                if trade['type'] != 'trade':
                    continue

                if str(trade['base']).upper() != self.tso.upper():
                    continue

                if trade['asset'] not in [self.symbol]:
                    continue

                if abs(pendulum.now().diff(pendulum.parse(trade['timestamp_at_exchange'])).in_seconds()) > 30:
                    continue

                # timestamp = int(pendulum.parse(trade['timestamp_at_exchange']).timestamp())
                timestamp = int(pendulum.parse(trade['timestamp_at_connector']).timestamp())
                if timestamp not in Trades.get_instance().trades:
                    Trades.get_instance().trades[timestamp] = []

                Trades.get_instance().trades[timestamp].append(trade)
                SharedData.get_instance().last_close = trade['price']

                # print(f"Status {self.cursor.status} {self.asset}".upper())
                if self.cursor.status == CursorStatus.PENDING:
                    self.cursor.status = CursorStatus.STARTED

                if timestamp not in self.cursor.ticks:
                    print(f"Creating Timestamp Not Available {timestamp} {trade['timestamp_at_exchange']} {self.symbol}".upper())
                    self.cursor.ticks[timestamp] = Tick(symbol=self.symbol, stamp=timestamp)

                self.input_trading_data_queue.task_done()
        except Exception as e:
            raise e



import asyncio
from typing import Optional

from komoutils.core import KomoBase
from pydantic import BaseModel

from aporacle.intervaled_data_symbol.client.orchestrator import Orchestrator


class AssetPriceData(BaseModel):
    chain: str
    epoch: int
    asset: str
    price: float


class IntervaledTradingSymbolExecutor(KomoBase):

    def __init__(self,
                 symbol: str,
                 input_queue: asyncio.Queue,
                 output_queue: asyncio.Queue,
                 output_record_evaluation_ticks: int = 18
                 ):
        super().__init__()
        self.symbol: str = symbol
        self.asset: str = self.symbol.split("_")[-2]
        self.input_queue: asyncio.Queue = input_queue
        self.output_queue: asyncio.Queue = output_queue
        self.output_record_evaluation_ticks = output_record_evaluation_ticks
        self.orchestrator: Optional[Orchestrator] = None

    @property
    def name(self):
        return "intervaled_trading_symbol_executor"

    def start(self):
        self.orchestrator = Orchestrator(
            symbol=self.symbol,
            input_queue=self.input_queue,
            output_queue=self.output_queue,
            output_record_evaluation_ticks=self.output_record_evaluation_ticks,
        )
        self.orchestrator.start()

    def stop(self):
        pass

    def set_tso_data(self, asset_price_data: AssetPriceData):
        self.orchestrator.tso_price_set(chain=asset_price_data.chain,
                                        tso=asset_price_data.asset,
                                        price=asset_price_data.price)

    def process_chain_epoch(self, chain: str, epoch: int):
        return self.orchestrator.generate_chain_epoch_record(chain=chain, epoch=epoch)
